﻿/*
 * Licensed under the EUPL, Version 1.1 or – as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence. You may
 * obtain a copy of the Licence at:
 * 
 * http://www.osor.eu/eupl/european-union-public-licence-eupl-v.1.1
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * Licence for the specific language governing permissions and limitations under
 * the Licence.
 */
using System;
using System.Xml;
using System.Security.Cryptography.X509Certificates;
using System.Security.Cryptography.Xml;
using System.Security.Cryptography;

using eu.stork.peps.auth.commons.Exceptions;

//
// author: AMA – Agência para a Modernização Administrativa IP, PORTUGAL (www.ama.pt)
//
namespace eu.stork.peps.auth.commons
{
    /// <summary>
    /// This class contains signature related methods
    /// </summary>
    public class SignatureUtils
    {
        /// <summary>
        /// The XML namespace of XmlDSig
        /// </summary>
        public const string XMLDSIG = "http://www.w3.org/2000/09/xmldsig#";

        /// <summary>
        /// Verifies if a signature within an xml document is valid
        /// </summary>
        /// <param name="doc">an xml document with a signature element</param>
        /// <returns>true if the signature is valid; false otherwise</returns>
        public static bool VerifySignature(XmlDocument doc)
        {
            try
            {
                if (!doc.PreserveWhitespace)
                    return false;
                return VerifySignature(doc.DocumentElement);
            }
            catch (Exception ex)
            {
                // TODO logging
                throw new SignatureUtilsException("Exception occurred on SignatureUtils.VerifySignature", ex);
            }
        }

        /// <summary>
        /// Verifies if a signature within an xml element is valid
        /// </summary>
        /// <param name="doc">an xml document with a signature element</param>
        /// <returns>true if the signature is valid; false otherwise</returns>
        public static bool VerifySignature(XmlElement element)
        {
            try
            {
                SignedXml signedXml = new SignedXml(element);
                XmlNodeList nodeList = element.GetElementsByTagName("Signature", XMLDSIG);
                if (nodeList.Count == 0)
                    return false;
                signedXml.LoadXml((XmlElement)nodeList[0]);
                return signedXml.CheckSignature();
            }
            catch (Exception ex)
            {
                // TODO logging
                throw new SignatureUtilsException("Exception occurred on SignatureUtils.VerifySignature", ex);
            }
        }

        /// <summary>
        /// Given a certificate, this method verifies if the signature within an xml document is valid or not
        /// </summary>
        /// <param name="el">an xml document element</param>
        /// <param name="certPub">a certificate containing the public key to be used on the signature</param>
        /// <returns>true if the signature is valid; else otherwise</returns>
        public static bool VerifySignature(XmlElement el, X509Certificate2 certPub)
        {
            try
            {
                SignedXml signedXml = new SignedXml(el);

                XmlNodeList nodeList = el.GetElementsByTagName("Signature", XMLDSIG);
                if (nodeList.Count == 0) // Document does not contain a signature to verify
                    // TODO error code?
                    return false;

                using (RSACryptoServiceProvider RSA2 = (RSACryptoServiceProvider)certPub.PublicKey.Key)
                {
                    RSAParameters KeyPub = RSA2.ExportParameters(false);
                    signedXml.LoadXml((XmlElement)nodeList[0]);
                    return signedXml.CheckSignature(RSA2);
                }
            }
            catch (Exception ex)
            {
                // TODO logging
                throw new SignatureUtilsException("Exception occurred on SignatureUtils.VerifySignature", ex);
            }
        }

        /// <summary> 
        /// Signs an xml document and attaches the signature element after the reference node.
        /// </summary>
        /// <param name="doc">an xml document to be signed</param>
        /// <param name="cert">a certificate containing a private key</param>
        /// <param name="refNode">the reference node, the signature will be attached after it</param>
        public static void SignDocument(XmlDocument doc, X509Certificate2 cert, XmlNode refNode)
        {
            try
            {
                SignDocument(doc, "", cert, refNode);
            }
            catch (Exception ex)
            {
                // TODO logging
                throw new SignatureUtilsException("Exception occurred on SignatureUtils.SignDocument", ex);
            }
        }

        /// <summary> 
        /// Signs an xml element and attaches the signature element after the reference node.
        /// </summary>
        /// <param name="doc">an xml document to be signed</param>
        /// <param name="cert">a certificate containing a private key</param>
        /// <param name="refNode">the reference node, the signature will be attached after it</param>
        public static void SignDocument(XmlElement element, X509Certificate2 cert, XmlNode refNode)
        {
            try
            {
                SignDocument(element, "", cert, refNode);
            }
            catch (Exception ex)
            {
                // TODO logging
                throw new SignatureUtilsException("Exception occurred on SignatureUtils.SignDocument", ex);
            }
        }

        /// <summary>
        /// Signs an xml document with a given id and attaches the signature after
        /// a given reference node.
        /// </summary>
        /// <param name="doc">an xml document</param>
        /// <param name="id">the id of an element to be signed</param>
        /// <param name="cert">a certificate with a private key</param>
        /// <param name="refNode">a reference node</param>
        public static void SignDocument(XmlDocument doc, string id,
            X509Certificate2 cert, XmlNode refNode)
        {
            try
            {
                SignDocument(doc.DocumentElement, id, cert, refNode);
            }
            catch (Exception ex)
            {
                // TODO logging
                throw new SignatureUtilsException("Exception occurred on SignatureUtils.SignDocument", ex);
            }
        }

        /// <summary>
        /// Signs an xml document with a given id and attaches the signature after
        /// a given reference node.
        /// </summary>
        /// <param name="doc">an xml document</param>
        /// <param name="id">the id of an element to be signed</param>
        /// <param name="cert">a certificate with a private key</param>
        /// <param name="refNode">a reference node</param>
        public static void SignDocument(XmlElement element, string id,
            X509Certificate2 cert, XmlNode refNode)
        {
            try
            {
                SignedXml signedXml = new SignedXml(element);
                signedXml.SigningKey = cert.PrivateKey;

                Reference reference = new Reference(id == String.Empty ? id : "#" + id);
                reference.AddTransform(new XmlDsigEnvelopedSignatureTransform());
                reference.AddTransform(new XmlDsigExcC14NTransform());
                signedXml.AddReference(reference);
                signedXml.KeyInfo.AddClause(new KeyInfoX509Data(cert));
                signedXml.ComputeSignature();
                XmlElement xmlDigitalSignature = signedXml.GetXml();
                element.InsertAfter(xmlDigitalSignature, refNode);

            }
            catch (Exception ex)
            {
                // TODO logging
                throw new SignatureUtilsException("Exception occurred on SignatureUtils.SignDocument", ex);
            }
        }

    }
}
